; *****************************************************
; ZX80 (8K ROM) / ZX81 1K Space Invaders - Main Program
; *****************************************************
; (c)1981 Macronics
;
; RAM required: 1K
;
; Disassembled by Paul Farrow, www.fruitcake.plus.com
; Initial version 17th April 2011 (revised 24th April 2011)


; =====
; Notes
; =====

; The program spans from $4000 (16384) to $41EF (16879) inclusive.
;
; The Display file is located at $41FF (16895). It begins with a newline character ($76), followed by 16 rows of
; 31 columns. Each column is terminated with a newline character ($76).
;
; The following program variables are used:
;
; IY+$1B        $41F5/F6 = Top of stack. $41F6 holds a sequence counter which indicates what action to perform at each frame.
;                          The bit combinations have the following meanings:
;                            00 = Move a row of invaders.
;                            01 = Move the invader bomb and laser base missile.
;                            10 = Read keyboard and move laser base.
;                            11 = Check whether to drop a new invader bomb.
; IY+$1C/1D     $41F7/F8 = Invader bomb display file position or $00xx for none.
; IY+$1E/1F     $41F9/FA = Laser base missile display file position or $00xx for none.
; IY+$20        $41FB    = Holds the number of invaders currently on screen (up to a maximum of 16).
; IY+$21        $41FC    = A flag used to indicate whether an invader has been selected to drop a bomb ($FF=Invader selected, else $00).
; IY+$22        $41FD    = Low address byte of the laser base display file position.
; IY+$23        $41FE    = Holds the number of scan lines to produce. It is used by the ROM display routines, which addresses this
;                          location using IY+$23. [This value is only applicable in the ZX80 version of the game and even then might not have been used]
;               $41FF    = Display file, ending at $43FF.


; ===============
; ZX80 Conversion
; ===============

; The program can be modified to assemble for running on a ZX80 by uncommenting the following line:

;#define ZX80

; Macronics sold a number of programs that were available for both the ZX80 and ZX81. They achieved this by keeping the game code the same
; and using a custom video driver that generated the top and bottom borders in software, i.e. replicated the manner in which the ZX81 produces
; its display when in FAST mode. As a result, the same program could be used on the ZX80 fitted with the upgrade ROM. With only small changes
; to the video driver and some of the character set codes, the program could also be run on the ZX80 fitted with its original ROM. By comparing
; the differences used in their other games available for both machines, it is possible to make a conjecture on how the ZX80 version of
; 1K Space Invaders would have been.

#ifndef ZX80

; ---- ZX81 VERSION ----

#define CH_INVADER_LEFT         $87     ; Left character of an invader.
#define CH_INVADER_MIDDLE       $0B     ; Middle character of an invader.
#define CH_INVADER_RIGHT        $04     ; Right character of an invader.
#define CH_INVADER_BOMB         $97     ; Character for an invader bomb.

#else

; ---- ZX80 VERSION ----

#define CH_INVADER_LEFT         $07     ; Left character of an invader.
#define CH_INVADER_MIDDLE       $01     ; Middle character of an invader.
#define CH_INVADER_RIGHT        $06     ; Right character of an invader.
#define CH_INVADER_BOMB         $94     ; Character for an invader bomb.

#endif


; ====================
; Assembler Directives
; ====================

; For use with the TASM assembler.

#define DEFB .BYTE
#define DEFW .WORD
#define DEFM .TEXT
#define DEFS .FILL
#define EQU  .EQU
#define ORG  .ORG
#define END  .END

; ======================================================================================================================================================

        ORG     $4000

; -------------------
; Program Entry Point
; -------------------

L4000:  LD      IY,$41DB                ; Base address of program variables.

; The display file is populated by setting the stack to point at the display file and then
; pushing the data into memory. Since the stack grows downwards, the display file is set up
; in reverse order.

        LD      SP,$43FF                ; Point to the end of the display file.

; Load the score characters.

        LD      DE,$1C1C                ; DE holds "00" forming the digits of the score.
        PUSH    DE                      ; Load the score characters to the bottom right of the display file (addresses $43FD/E).

; Load the remainder of the next 3 rows with inverse spaces.

        LD      DE,$8080                ; DE holds two inverse spaces.
        LD      B,$2F                   ; 47*2=94 inverse spaces.

L4010:  PUSH    DE                      ; Load the inverse spaces into memory.
        DJNZ    L4010                   ; Repeat until all 3 rows have been populated.

        LD      HL,$0000                ; HL holds two space characters which are used to represent to shields.
        LD      B,$10                   ; 4 rows of 4 shields.

L4018:  PUSH    DE                      ; Insert two inverse spaces.
        PUSH    HL                      ; Insert two spaces, i.e. shield characters.
        PUSH    HL                      ; Insert two spaces, i.e. shield characters.
        PUSH    DE                      ; Insert two inverse spaces.
        DJNZ    L4018                   ; Repeat until all 4 rows have been populated.

; Fill the remaining 9 rows with inverse spaces.

        LD      B,$90                   ; 144*2=288, i.e. 9 rows.

L4020:  PUSH    DE                      ; Load the inverse spaces into memory.
        DJNZ    L4020                   ; Repeat until all 9 rows have been populated.

; Initialise program variables.

        LD      DE,$50CB                ; There are 80 ($50) lines in the top border [used in ZX80 version only if calling $01AD], and the laser base display file address is $43CB (14,11).
                                        ; [Change this to 56 ($38) scan lines for USA 60Hz]
        PUSH    DE
        PUSH    HL                      ; Clear all other program variables to $00.
        PUSH    HL
        PUSH    HL
        PUSH    HL

; Insert a newline character at the start of the display file and at the end of each row, i.e. as the 32nd character of each row.

        LD      DE,$0020                ; Offset to the next display file row.
        LD      HL,$41FF                ; Point to the byte before the display file.
        LD      B,$11                   ; 17 characters, i.e. there are 16 rows and an initial terminator.

L4033:  LD      (HL),$76                ; Insert the row terminator.
        ADD     HL,DE                   ; Advance to the next row.
        DJNZ    L4033                   ; Repeat for all rows.

; Draw the last score at the bottom left hand corner.

        LD      HL,($40E4)              ; Fetch the last score characters.
        LD      ($43E0),HL              ; Load the last score into the display file.

; The main program loop used to commence a new wave of invaders.

L403E:  LD      HL,$42FD                ; Set the address of the next row within the invader area of the display file to start shifting from as row 7, column 29.
        LD      C,$08                   ; The number of shifts remaining before the invaders reached the edge of the display and will require moving down a row. B will hold $00.
        EXX

; Draw the two rows of space invaders.

        LD      A,$02                   ; There are 2 rows to draw.
        LD      HL,$4221                ; Row 7, column 1.

L4049:  LD      B,$04                   ; There are 4 invaders per row.

L404B:  LD      (HL),CH_INVADER_LEFT    ; Draw the left character of an invade [$87=ZX81, $07=ZX80]r.
        INC     L
        LD      (HL),CH_INVADER_MIDDLE  ; Draw the middle character of an invader [$0B=ZX81, $01=ZX80].
        INC     L
        LD      (HL),CH_INVADER_RIGHT   ; Draw the right character of an invader [$04=ZX81, $06=ZX80].

        INC     L                       ; Advance to the location for the next invader.
        INC     L
        INC     L
        INC     L
        DJNZ    L404B                   ; Loop back to draw all invaders on the current row.

        LD      L,$61                   ; Row 9, column 1.
        DEC     A                       ; Decrement the row count.
        JR      NZ,L4049                ; Jump back if there is another row of invaders to draw.

        LD      (IY+$20),$08            ; Set the number of invaders on screen (initially there are 2 rows of 4 invaders, i.e. 8 in total).

; A PAL TV picture consists of 25 frames per second, consisting of 625 scan lines. Each frame is divided into 2 interlaced fields, referred to
; as odd and even fields, with each field consisting of 312.5 scan lines. A field occupies 20ms and hence each scan line takes 64us.
; To produce a steady TV picture, the FAST mode display routine within the ZX81 ROM has been carefully structured so that each iteration of the loop
; takes exactly 20ms to execute. The ZX81 is driven by a 6.5MHz ceramic resonator/filter, which is halved and tuned down to approximately 3.21MHz.
; Each iteration of display loop within the ROM when in FAST mode takes 64163 T-states and this spans 20ms. Therefore each T-states lasts 311.71ns,
; and hence the CPU clock frequency assumed by the ZX81 is 3.20815MHz . Ceramic resonators do not have a tight tolerance and so some variation
; around this figure is to be expected. The ZX81 ROM outputs 303 scan lines (56 in the top border, 192 in the main area and 55 in the bottom border),
; with each requiring 207 T-states to generate. Therefore 303 * 207 = 62721 T-states are required to span these scan lines, with an additional
; 73 T-states used to return control to the display loop. The remaining 1369 T-states are used to read the keyboard and update the FRAMES system
; variable. Note that with a clock frequency of 3.20815MHz, each scan line takes 64.52us. This means that the TV field produced by the ZX81 spans
; 309.97 of these slightly longer scan lines.
;
; The 1K Space Invaders game contains a display loop that outputs 288 scan lines per TV field (80 in the top border, 128 in the main display area,
; and 80 in the bottom border). This frees up 15 * 207 = 3105 T-states, making a total of 4474 T-states (equating to 1.395ms) available per TV field
; for running the game. To achieve a steady display, the functionality performed at each field must be structured to fill this time as accurately
; as possible. The game code is divided into four sections, and these are cycled through one per TV field. It takes 83 T-states to run the sequence
; control loop, leaving 4391 T-states (1.369ms) for each game section. The game sections actually take the following durations:
;
; Part 1 = 4474 T-states  -  Move row of invaders.
; Part 2 = 4460 T-states  -  Move invader bomb and laser base missile.
; Part 3 = 4467 T-states  -  Read keyboard and move laser base.
; Part 4 = 4468 T-states  -  Check whether to drop new invader bomb.
;
; These figures are too large. It appears that to compensate for the 83 T-states it takes to run the sequence control loop, the time to produce the first
; scan line was reduced by 89 T-states (see below). However, this is after the vertical sync generator has been switched on and so results in the first scan line
; of the display being shifted to the left hand edge of the TV picture. Subsequent scan lines are then produced correctly. The first scan line is the start of
; the top border and so is pure white. Even if this scan line were within the visible area of a TV set, it would not be evident that the line was actually shifted
; to the left.
;
; There is a range of values for each section due to small differences that can occur between different paths within those routines. The values
; are all close enough to the target value of 1.395ms that any difference simply causes the next field to be displayed slightly sooner or later
; than the desired 20ms rate. Although it is desirable to produce TV fields at a precise rate of 20ms, a small amount of variation is tolerable
; and television sets can cope with this. The target duration of 1.395ms is based on the CPU speed assumed by the ZX81 ROM, but this may vary
; slightly between machines due to tolerance of the resonator and so is only a guide to the time available.

; ---- ZX80 VERSION ----
; A PAL TV picture consists of 25 frames per second, consisting of 625 scan lines. Each frame is divided into 2 interlaced fields, referred to
; as odd and even fields, with each field consisting of 312.5 scan lines. A field occupies 20ms and hence each scan line takes 64us.
; To produce a steady TV picture, the display routine within the ZX80 ROM has been carefully structured so that each iteration of the loop
; takes exactly 20ms to execute. The ZX80 is driven by a 6.5MHz ceramic resonator/filter, which is halved and tuned down to approximately 3.21MHz.
; Each iteration of display loop within the ROM takes 64167 T-states and this spans 20ms. Therefore each T-states lasts 311.67ns, and hence the
; CPU clock frequency assumed by the ZX80 is 3.20835MHz (which is close to the 3.2MHz marked on the ZX80 schematic). Ceramic resonators do not
; have a tight tolerance and so some variation around this figure is to be expected. The ZX80 ROM outputs 303 scan lines (56 in the top border,
; 192 in the main area and 55 in the bottom border), with each requiring 207 T-states to generate. Therefore 303 * 207 = 62721 T-states are
; required to span these scan lines, with an additional 27 T-states used to return control to the display loop. The remaining 1419 T-states
; are used to read the keyboard and update the FRAMES system variable. Note that with a clock frequency of 3.20835MHz, each scan line takes 64.52us.
; This means that the TV field produced by the ZX80 spans 309.99 of these slightly longer scan lines.
;
; The 1K Space Invaders game contains a display loop that outputs 288 scan lines per TV field (80 in the top border, 128 in the main display area,
; and 80 in the bottom border). This frees up 15 * 207 = 3105 T-states, making a total of 4524 T-states (equating to 1.410ms) available per TV field
; for running the game. To achieve a steady display, the functionality performed at each field must be structured to fill this time as accurately
; as possible. The game code is divided into four sections, and these are cycled through one per TV field. It takes 83 T-states to run the sequence
; control loop, leaving 4441 T-states (1.384ms) for each game section. The game sections actually take the following durations:
;
; Part 1 = 4474 T-states  -  Move row of invaders.
; Part 2 = 4460 T-states  -  Move invader bomb and laser base missile.
; Part 3 = 4467 T-states  -  Read keyboard and move laser base.
; Part 4 = 4468 T-states  -  Check whether to drop new invader bomb.
;
; These figures are slightly too large but close enough to be tolerated by a TV.
;
; There is a range of values for each section due to small differences that can occur between different paths within those routines. The values
; are all close enough to the target value of 1.410ms that any difference simply causes the next field to be displayed slightly sooner or later
; than the desired 20ms rate. Although it is desirable to produce TV fields at a precise rate of 20ms, a small amount of variation is tolerable
; and television sets can cope with this. The target duration of 1.410ms is based on the CPU speed assumed by the ZX80 ROM, but this may vary
; slightly between machines due to tolerance of the resonator and so is only a guide to the time available.
; ---- ZX80 VERSION ----

L4062:  LD      DE,$0020                ; 10    Holds the offset to the next row.
        LD      C,$80                   ; 7     Holds the code for an inverse space.

        IN      A,($FE)                 ; 11    Turn on the vertical sync generator. It be will be turned off approximately 23 scan lines from now.
                                        ;       [Any I/O read to an address with A0 low will turn on the vertical sync generator]

        POP     AF                      ; 10    Fetch the byte at $41F6, which is used as a sequence counter to determine what action to perform this frame.
        INC     A                       ; 4     Increment the counter.
        PUSH    AF                      ; 11    Store the updated counter.

        AND     $03                     ; 7     Keep only the lower 2 bits, forming 4 possibilities to consider.
        JR      Z,L4096                 ; 12/7  Move a row of invaders (typical 3342 T-states, returning with delay counter B=$56).

        DEC     A                       ; 4     Sequence count of 1?
        JR      Z,L40EF                 ; 12/7  Move the invader bomb and laser base missile (typically 1122 T-states, returning with delay counter B=$00).

        DEC     A                       ; 4     Sequence count of 2?
        JP      Z,L4166                 ; 10/10 Read the keyboard and move the laser base (typically 1119 T-states, returning with delay counter B=$00).

        DEC     A                       ; 4     Sequence count of 3?
        JP      Z,L41AD                 ; 10/10 Check whether to drop a new invader bomb (typically 1259 T-states, returning with delay counter B=$F3).
                                        ;       [This jump will always be taken if the CPU has reached this point]

; This is the return point from all four processing routines.

L407B:  DJNZ    L407B                   ; 13/8  Perform a delay to pad out until the end of the TV field.

; With the delay above, the total (average) time taken by each processing routine is:
;
; Part 1 = 4557 T-states  -  Move row of invaders.
; Part 2 = 4542 T-states  -  Move invader bomb and laser base missile.
; Part 3 = 4550 T-states  -  Read keyboard and move laser base.
; Part 4 = 4552 T-states  -  Check whether to drop new invader bomb.
;
; Generate the top TV border and main area.

#ifndef ZX80

; ---- ZX81 VERSION ----

        OUT     ($FD),A                 ; 11    Turn off the vertical sync generator.

; The first scan line produced by the program is too short but this does not cause a problem for TV sets since all subsequent scan lines are produced correctly.
; The first scan line is too short by 37 T-states. The problem can be improved by loading a value of $E6 into the A register instead of $EF. This will introduce
; an additional 36 T-states, and so reduces the error to -1 T-state. The A register holds a counter that is incremented until it reaches 0, at which point the start
; of the next scan line is triggered. Each decrement of the value loaded into the A register introduces a 4 T-states delay, and hence decrementing the value by 9
; introduces a 28 T-states delay.
;
; In FAST mode, the ZX81 uses software to generate the top and border borders, and this is what the routine below also does. In FAST mode, the ZX81 ROM takes 120 T-states
; from issuing OUT ($FD),A to before the CALL $02B5. The routine below takes 31 T-states, and hence is 89 T-states too short. The ROM loads the A register with $FC where as
; the routine below uses a value of $EF in an attempt to compensate for the deficit. The difference of $FC-$EF=$0D introduces 13*4=52 T-states, reducing the deficit to
; 89-52=37 T-states. If the A register is loaded with $E6 then this would introduce a 9*4=36 T-state delay, thereby reducing the deficit to 1 T-state.
;
; The 1K Space Invaders game is written to generate a 50Hz display field rate suitable for the UK. The ZX81 ROM generates a 60Hz display field rate
; by reducing the number of top and bottom border lines by 24 each. The game can be modified to support a 60Hz display rate by modifying the
; the two lines below and the one above as indicated.

        LD      A,$EF                   ; 7     Timing counter for the time left to generate for the first scan line of the top border. Thereafter set to $DD by the ROM.
        LD      B,$11                   ; 7     There are 17 HALTs within the display file, i.e. 16 rows and 1 top border.
        LD      HL,$C1FF                ; 10    Point to the start of the display file, i.e. the initial HALT representing the top border, but with bit 15 of the address set.
        LD      C,$50                   ; 7     There are 80 scan lines in the top border.
                                        ;       [Change this to 56 ($38) scan lines for USA 60Hz]
        CALL    $02B5                   ; 17    Generate the top border and main display area.

; Generate the bottom TV border.

; The ZX81 ROM would take 105 T-states after returning from CALL $02B5 to before the next CALL $02B5. This routine takes 24 T-states, and
; hence is 81 T-states too short. The ZX81 ROM loads the A register with $FC and so this routine loads it with $E8 to compensate for the deficit.
; This difference of $FC-$E8=$14 introduces 20x4=80 T-states and hence reduces the deficit to -1 T-state.

        LD      A,$E8                   ; 7     Timing counter for the time left to generate for the first scan line of the bottom border.
        INC     B                       ; 4     Increment the row count, i.e. B holds $01.
        DEC     HL                      ; 6     Decrement the screen address, i.e. point to the last HALT.
        LD      C,$50                   ; 7     There are 80 scan lines in the top border.
                                        ;       [Change this to 56 ($38) scan lines for USA 60Hz]
        CALL    $02B5                   ; 17    Generate the bottom border.

#else

; ---- ZX80 VERSION ----

        OUT     ($FE),A                 ; 11    Turn off the vertical sync generator.

; The timing counter values used for the top and bottom borders are based on those used in 1K Breakout for the ZX80. This game uses values of $EC and $F0 respectively,
; and so by compensating for the difference in instructions used results in values of $EE and $EF.
;
; The first scan line produced by the program is too short but this does not cause a problem for TV sets since all subsequent scan lines are produced correctly.
; The first scan line is too short by 15 T-states. The problem can be resolved by loading a value of $EA into the A register instead of $EE. This will introduce
; an additional 16 T-states, and so reduces the error to +1 T-state. The A register holds a counter that is incremented until it reaches 0, at which point the start
; of the next scan line is triggered. Each decrement of the value loaded into the A register introduces a 4 T-states delay, and hence decrementing the value by 4
; introduces a 16 T-states delay.
;
; The ZX80 ROM takes 38 T-states from issuing OUT ($FF),A to before the CALL $01AD. The routine below takes 31 T-states, and hence is 7 T-states too short.
; The ROM loads the A register with $EC where as the routine below uses a value of $EE and so is 2*4=8 T-states faster. This results in a total deficit of 15 T-states.
; If the A register is loaded with $EA then this would introduce a 4*4=16 T-state delay, thereby reducing the deficit to +1 T-state.
;
; The 1K Space Invaders game is written to generate a 50Hz display field rate suitable for the UK. The ZX80 ROM generates a 60Hz display field rate
; by reducing the number of top and bottom border lines by 24 each. The game can be modified to support a 60Hz display rate by modifying the
; the two lines below and the one above as indicated.

        LD      A,$EE                   ; 7     Timing counter for the time left to generate for the first scan line of the top border. Thereafter set to $DD by the ROM.
        LD      B,$11                   ; 7     There are 17 HALTs within the display file, i.e. 16 rows and 1 top border.
        LD      HL,$C1FF                ; 10    Point to the start of the display file, i.e. the initial HALT representing the top border, but with bit 15 of the address set.
        LD      C,$50                   ; 7     This is harmless if a call to $01AD is made, and is a remnant from the ZX81 version.
                                        ;       It is only used if a call to $01B0 is made instead. [Change this to 56 ($38) scan lines for USA 60Hz]
        CALL    $01AD                   ; 17    Generate the top border and main display area. [Might have been CALL $01B0]

; Generate the bottom TV border.

; In ZX80 1K Breakout, a timing value of $F0 is used. However, these instructions do not contain a DEC (IY+$23) but instead have a LD C,$50.
; This reduces the execution time by 16 T-states and so to compensate the timing value is adjusted by 4 to $EC.
;
; The ZX80 ROM would take 52 T-states after returning from CALL $01AD to before the next CALL $01AD. This routine takes 24 T-states, and
; hence is 28 T-states too short. The ZX80 ROM loads the A register with $F3 and so this routine loads it with $EC to compensate for the deficit.
; This difference of $F3-$EC=$07 introduces 7x4=28 T-states and hence reduces the deficit to 0 T-states.

        LD      A,$EC                   ; 7     Timing counter for the time left to generate for the first scan line of the bottom border.
        INC     B                       ; 4     Increment the row count, i.e. B holds $01.
        DEC     HL                      ; 6     Decrement the screen address, i.e. point to the last HALT.
        LD      C,$50                   ; 7     This is harmless if a call to $01AD is made, and is a remnant from the ZX81 version.
                                        ;       It is only used if a call to $01B0 is made instead. [Change this to 56 ($38) scan lines for USA 60Hz]
        CALL    $01AD                   ; 17    Generate the bottom border. [Might have been CALL $01B0]
#endif

        JR      L4062                   ; 12    Repeat the main program loop for the current wave of invaders.

; -------------
; Move Invaders
; -------------
; At each invocation, this routine shifts all characters on a particular row by one position left or right, thereby
; moving the invaders. All characters on a row are shifted each time, whether blank or not, and hence this ensures that
; the routine completes after a fixed duration of time. The area that the invaders can occupy spans rows 0 to 7.
; On the odd rows the invaders are shifted to the right, and on the even rows they are shifted to the left.
; Rows are shifted in decending order, beginning with row 7 and ending with row 0. Initially the invaders are located at the
; left hand side of the screen and so are shifted to the right. The invaders are shifted by eight columns before moving down a row.
; The invaders cannot occupy columns 0 or 31, and so these always filled with black space. This black space is shifted along with
; the invaders, thereby overwriting the 'trail' character left by an invader. The black space is not overwritten itself.
; Note that it is possible for the bottom row of invaders to be shifted onto row 8, where they will remain static until erased
; by shifting the blank row above over them. The game does not end when the invaders hit the shields and so it is possible for
; all invaders to move to row 8 and subsequently erased. There is then no way to reset the game, apart from power cycling the computer
; and reloading the game.
; Entry: HL'=The address of the next row within the invader area of the display file to start shifting from.
;        C'=The number of shifts remaining before the invaders reached the edge of the display and will require moving down.
;        B'=$00.
; The routine typically takes 3342 T-states and returns with a delay counter of $56 to yield a delay of 1113 T-states, making
; a total of 4455 T-states.

L4096:  EXX                             ; 4     Fetch invader area variables.

        LD      A,L                     ; 4     Point to the last display file address of the invaders.
        AND     $20                     ; 7     Extract the row address. Is it odd or even?
        JR      Z,L40A1                 ; 12/7  Jump if even to move invaders to the left.

; Move the invaders to the right.

        LD      DE,$FFFF                ; 10    A value of -1 will be subtracted, thereby moving the invaders to the right.
        JR      L40A5                   ; 12    Jump ahead to continue.

; Move the invaders to the left.

L40A1:  LD      DE,$0001                ; 10    A value of 1 will be subtracted, thereby moving the invaders to the left.

        LD      A,(HL)                  ; 7     Perform a delay.

L40A5:  LD      A,(HL)                  ; 7     Fetch a character from the invader area.

; Enter a loop to shift all characters on the current row.

L40A6:  INC     C                       ; 4     Is it time to move the invaders down to the next row?
        DEC     C                       ; 4
        JR      Z,L40B2                 ; 12/7  Jump ahead if so.

; It is not yet time to move the invaders down, so shift a character on the current row over.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Add the offset to generate the address to move the character to.
        LD      (HL),A                  ; 7     Shift the character.

        ADD     HL,DE                   ; 11    Point back to the original address.
        NOP                             ; 4     Perform a delay.
        JR      L40BA                   ; 12    Jump ahead to continue.

; It is time to move the invaders down.

L40B2:  SET     5,C                     ; 8     This sets C to hold 32, i.e. the offset to the row below.
        ADD     HL,BC                   ; 11    Generate the address of the next row.
        LD      (HL),A                  ; 7     Shift the invader character.

        SBC     HL,BC                   ; 15    Point back to the original row.

        LD      C,$00                   ; 7     Perform a delay.

; A character on the row has been shifted. Test whether all characters in the current row have been shifted.

L40BA:  ADD     HL,DE                   ; 11    Point to the next location to shift.
        LD      A,(HL)                  ; 7
        CP      $76                     ; 7     Has the end/start of the row been reached?
        JR      NZ,L40A6                ; 12/7  Jump back if not to move the next character on this row.

; All characters on the current row have been shifted.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point back to the last character shifted within the row.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point to the edge position within the row.

        LD      DE,$FFE0                ; 10    Offset of -32.
        ADD     HL,DE                   ; 11    Point to the row above.

        LD      A,H                     ; 4     Has the top of the display file been exceeded?
        CP      $41                     ; 7
        JR      NZ,L40DB                ; 12/7  Jump ahead if not, i.e. there are further rows to be shifted next time.

; The whole of the invader area has been shifted, i.e. one cycle has been completed.

        LD      HL,$42FD                ; 10    Reset the next display file position to shift back to row 7, column 29.
        INC     C                       ; 4
        DEC     C                       ; 4     Have the invaders been moved to the edge of the display?
        JR      NZ,L40D8                ; 12/7  Jump ahead if not. Following this jump results in a quicker path by 2 T-states.

; The invaders have been shifted enough times such that they are now at the edge of the display.

        LD      C,$09                   ; 7     Reset the number of shifts, including one to cancel out the immediate decrement below.

L40D8:  DEC     C                       ; 4     Decrement the count of the number of columns the invaders have been shifted.
        JR      L40DF                   ; 12    Jump ahead to continue.

; A complete row of the invader area has been shifted.

L40DB:  LD      B,$03                   ; 7     Perform a delay to match the other paths through this routine.

L40DD:  DJNZ    L40DD                   ; 13/8

L40DF:  EXX                             ; 4     Save the invader area variables.

; Make a return to the main loop, executing a delay upon return to pad out the processing time slot.

        LD      B,$56                   ; 7     Set the delay counter.

L40E2:  JR      L407B                   ; 12    Return to the main loop.

; ----------------
; Last Score Store
; ----------------
; Holds the digits of the last score.

L40E4:  DEFB    $1C                     ; '0'. Most significant digit.
        DEFB    $1C                     ; '0'. Least signifcant digit.

; ----------
; Save Score
; ----------

L40E6:  LD      HL,($43FD)              ; 16    Fetch the current score characters.
        LD      ($40E4),HL              ; 16    Copy the score characters into the last score store.
        JP      L4000                   ; 10    Restart the game.

; --------------------------------------------
; Move the Invader Bomb and Laser Base Missile
; --------------------------------------------
; The routine typically takes 1122 T-states and returns with a delay counter of $100 to yield a delay of 3323 T-states, making
; a total of 4445 T-states.
; Entry: DE=Offset to the next row.
;        C=Inverse space.

; First move the invader bomb if there is one.

L40EF:  LD      HL,($41F7)              ; 16    Fetch the invader bomb display file position.
        LD      A,H                     ; 4     Fetch the high address byte.
        AND     A                       ; 4     Is there an active bomb?
        ADD     HL,DE                   ; 11    Address of the next row, where the bomb will move to.
        LD      A,H                     ; 4     Delay.
        JR      Z,L40FA                 ; 12/7  Jump ahead if there is not.

; There is an active invader bomb. Check whether it has reached row 16. If it has then its location
; within the display file will be at $4400 or greater.

        CP      $44                     ; 7     Has the invader bomb reached row 16?

L40FA:  JR      Z,L4108                 ; 12/7  Jump ahead if so.

        LD      A,(HL)                  ; 7     Fetch the character at the new invader bomb position.
        AND     A                       ; 4     Has it hit a shield?
        JR      Z,L410B                 ; 12/7  Jump ahead if so.

        CP      $0D                     ; 7     Is it '$', i.e. the laser base?
        JR      Z,L40E6                 ; 12/7  Jump if so to reset the score and start a new game.

        LD      (HL),CH_INVADER_BOMB    ; 10    Draw the invader bomb '*'. [$97=ZX81, $94=ZX80]
        JR      L410F                   ; 12    Jump ahead.

; The invader bomb has not reached row 16.

L4108:  ADD     HL,DE                   ; 11    Delay
        JR      L410C                   ; 12    Jump ahead.

; The invader bomb has hit a shield.

L410B:  LD      (HL),C                  ; 7     Remove a character from the shield.

L410C:  LD      H,D                     ; 4     Perform a delay to balance the alternate path to $410F.
        PUSH    BC                      ; 11
        POP     BC                      ; 10

; The invader bomb has been moved. four paths join here, each approximately taking the same amount of time to arrive:
;   Path 1 - There is not an active invader bomb (113 T-states).
;   Path 2 - The invader bomb has reached row 16 (111 T-states).
;   Path 3 - The invader bomb has hit a shield (106 T-states).
;   Path 4 - The invader bomb has not hit the laser base (112 T-states).

L410F:  LD      ($41F7),HL              ; 16    Store the new invader bomb display file position.

; Now move the laser base missile if there is one.

        LD      DE,$FFE0                ; 10    Offset to the previous row.
        LD      HL,($41F9)              ; 16    Fetch the laser base missile display file position.
        LD      A,H                     ; 4     Fetch the high address byte.
        AND     A                       ; 4     Is there an active missile?
        ADD     HL,DE                   ; 11    Point to the row above the laser base missile.
        LD      A,H                     ; 4     Delay.
        JR      Z,L4120                 ; 12/7  Jump ahead if there is not.

; There is an active laser base missile. Check whether it has moved off the top of the screen. If it has then
; its location within the display file will be at $41FF or lower.

        CP      $41                     ; 7     Has the laser base missle moved off the top of the screen?

L4120:  JR      Z,L4131                 ; 12/7  Jump if the laser base missile has moved off the top of the screen.

; The laser base missile has not moved off the top of the screen.

        LD      A,(HL)                  ; 7     Fetch the character at the new missile location.
        CP      CH_INVADER_BOMB         ; 7     Is it an invader bomb '*'? [$97=ZX81, $94=ZX80]
        JR      Z,L4140                 ; 12/7  Jump ahead if it is.

; The laser base missile has not hit an invader bomb.

        AND     A                       ; 4     Has it hit a shield?
        JR      Z,L4144                 ; 12/7  Jump ahead if so.

; The laser base missile has not hit a shield or an invader bomb.

        CP      C                       ; 7     Is there a black space at the new laser base missile position?
        JR      NZ,L4147                ; 12/7  Jump ahead if not.

; The laser base missile has not hit an invader or a shield or an invader bomb.

        LD      (HL),$AE                ; 10    Draw the laser base missile, i.e. an inverse 'I'.
        JR      L4137                   ; 12    Jump ahead.

; The laser base missile has moved off the top of the screen.

L4131:  LD      B,$04                   ; 7     Perform a delay to balance the alternate path to $4135.

L4133:  DJNZ    L4133                   ; 13/8

; Joins here when the laser base missile has hit an invader bomb or a missile has hit an invader.

L4135:  LD      H,$00                   ; 7     This will result in the display file position of the laser base missile being marked as no active missile.

; Joins here when store the new location of the active missile.

L4137:  LD      ($41F9),HL              ; 16    Store the new laser base missile display file position.

; Perform a delay of 821 T-states before returning. This ensures that the routine approximately matches
; the durations of the other routines called from the main loop.

        LD      B,$3F                   ; 7

L413C:  DJNZ    L413C                   ; 13/8

L413E:  JR      L40E2                   ; 12    Indirectly return to the main loop. The delay counter will hold $00, i.e. 256 iterations.

; The laser base missile has hit an invader bomb. Note that it is possible for both an invader bomb and the laser base missile
; to be moved past each other if they happened to be adjacent to each other. If there is a black cell between them then they will
; both be moved to the same location and hence will cancel each other out.

L4140:  LD      (IY+$1D),$00            ; 19    Clear the display file position of the invader bomb, i.e mark it as no active bomb.

L4144:  LD      (HL),C                  ; 7     Remove a character from the shield or erase the invader bomb.
        JR      L4135                   ; 12    Jump to continue.

; The laser base missile has hit an invader and so the invader is removed from the display file.
; Since it is possible for the laser base missile to have hit the right most character of the invader, the display file
; address is decrement by 2 positions to ensure that the start of the invader is found. However, if the left most character
; was hit then the decrement by 2 will result in a display file address 2 positions to the left of the invaders. The program
; therefore blanks out 5 characters to ensure all of the invader has been removed.

L4147:  DEC     L                       ; 4     Point back 2 character positions to ensure that the left hand side of the invader is reached.
        DEC     L                       ; 4

        LD      B,$05                   ; 7     There are 5 character positions to erase.

L414B:  LD      (HL),C                  ; 7     Erase a character by writing an inverse space.
        INC     L                       ; 4     Point to the next character of the invader.
        DJNZ    L414B                   ; 13/8  Loop back to erase all characters of the invader.

; Increment the score.

        LD      HL,($43FD)              ; 16    Fetch the score characters.
        INC     H                       ; 4     Increment the low digit of the score.
        LD      A,H                     ; 4     Fetch the low digit of the score.
        CP      $26                     ; 7     Has '9' been surpassed?
        JR      NZ,L415B                ; 12/7  Jump ahead if not.

; A carry has occurred in the low digit of the score.

        LD      H,$1C                   ; 7     Set the low character back to '0'.
        INC     L                       ; 4     Increment the high digit of the score.
                                        ;       [Note that there is no check on the high digit of the score overflowing]

L415B:  LD      ($43FD),HL              ; 16    Save the new score characters.

        DEC     (IY+$20)                ; 23    Decrement the count of the number of invaders left on screen.
        JP      Z,L403E                 ; 10/10 Jump if all invaders have been destroyed to commence a new wave of invaders.

; There are still some invaders remaining in the current wave.

        JR      L4135                   ; 12    Jump to store the new laser base missile display file position.

; -------------------------------
; Move and Display the Laser Base
; -------------------------------
; This routine reads the keyboard and moves the laser base. It can move from column 2 to column 28 on row 14.
; It is moved to the right by key 8 or 3, and moved to the left by key 5 or 6, and key 0 or 1 will fire a missile.
; The routine typically takes 1119 T-states and returns with a delay counter of $100 to yield a delay of 3323 T-states, making
; a total of 4442 T-states.
; Entry: DE=Offset to the next row.
;        C=Inverse space.

L4166:  LD      L,(IY+$22)              ; 19    Fetch the low address of the laser base position within the display file.
        LD      H,$43                   ; 7     The high display file address of the laser base.
        LD      (HL),C                  ; 7     Erase the laser base.

; Upper byte selects keyboard row to read.
;
;          Bit0  Bit1  Bit2  Bit3  Bit4    Bit4  Bit3  Bit2  Bit1  Bit0
;          ----  ----  ----  ----  ----    ----  ----  ----  ----  ----
; $F7FE =    1     2     3     4     5       6     7     8     9     0   = $EFFE
; $FBFE =    Q     W     E     R     T       Y     U     I     O     P   = $DFFE
; $FDFE =    A     S     D     F     G       H     J     K     L    N/L  = $BFFE
; $FEFE =  SHIFT   Z     X     C     V       B     N     M     .   SPACE = $7FFE

        LD      B,$E7                   ; 7     Select both halves of the top keyboard row (1-5 and 6-0).
        LD      C,$FE                   ; 7     The keyboard I/O port.
        LD      A,L                     ; 4     Fetch the low byte of the laser base display file address.
        IN      D,(C)                   ; 12    Read the keyboard.

        BIT     2,D                     ; 8     Was key 8 (or 3) pressed?
        JR      NZ,L417E                ; 12/7  Jump ahead if not.

; Key 8 (or 3) was pressed so attempt to move to the right.

        CP      $DC                     ; 7     Is the laser base at column 28?
        JR      Z,L417C                 ; 12/7  Jump ahead if so.

; The laser base is not yet at the right hand edge of the screen so it can be moved.

        INC     L                       ; 4     Increment the laser base display file address, i.e. one cell to the right.

L417C:  JR      L4181                   ; 12    Jump ahead to continue.

; Key 8 (or 3) was not pressed. Delay to balance other paths through this routine and then test for other key presses.

L417E:  INC     (HL)                    ; 11    Perform a delay to balance the alternate paths to $4181.
        DEC     (HL)                    ; 11
        NOP                             ; 4

L4181:  BIT     4,D                     ; 8     Was key 5 (or 6) pressed?
        JR      NZ,L418C                ; 12/7  Jump ahead if not.

; Key 5 (or 6) was pressed so attempt to move to the left.

        CP      $C2                     ; 7     Is the laser base at column 2?
        JR      Z,L418A                 ; 12/7  Jump ahead if so.

; The laser base is not yet at the left hand edge of the screen so it can be moved.

        DEC     L                       ; 4     Decrement the laser base display file address, i.e. one cell to the left.

L418A:  JR      L418F                   ; 12    Jump ahead to continue.

; Key 5 (or 6) was not pressed. Delay to balance other paths through this routine and then test for other key presses.

L418C:  INC     (HL)                    ; 11    Perform a delay to balance the alternate paths to $418F.
        DEC     (HL)                    ; 11
        NOP                             ; 4

L418F:  LD      (IY+$22),L              ; 19    Store the new laser base display file address.
        LD      (HL),$0D                ; 10    Display the laser base as a '$' to indicate that there is an active missile.

        LD      A,($41FA)               ; 13    Is there an active laser base missile?
        AND     A                       ; 4
        JR      NZ,L41A3                ; 12/7  Jump if there is not.

; There is an active laser base missile, so there is no need to test for the fire key since it is only possible
; to fire one missile at a time.

        BIT     0,D                     ; 8     Was key 0 (or 1) pressed?
        JR      NZ,L41A5                ; 12/7  Jump ahead if not.

; Key 0 (or 1) was pressed so fire a missile.

        LD      ($41F9),HL              ; 16    Store the address within the display file of the missile.
        JR      L41A7                   ; 12    Jump ahead to continue.

; There is not an active laser bas missile. Delay to balance other paths through this routine.

L41A3:  LD      A,(HL)                  ; 7
        LD      A,(HL)                  ; 7

; Joins here when key 0 (or 1) was not pressed. Delay to balance other paths through this routine.

L41A5:  PUSH    BC                      ; 11
        POP     BC                      ; 10

; Joins here when a missile was fired. Delay to balance other processing slots.

L41A7:  LD      B,$40                   ; 7

L41A9:  DJNZ    L41A9                   ; 13/8

L41AB:  JR      L413E                   ; 12    Indirectly return to the main loop. The delay counter will hold $00, i.e. 256 iterations.

; -------------------------------------
; Check Whether to Drop an Invader Bomb
; -------------------------------------
; This routine searches the invader area from row 7 up to row 0 at the column where the laser base is.
; It searches for an invader that is directly above the laser base and sets a flag when one is found. Since
; all rows are checked, this flag ensures that the lowest invader found is the one that drops the bomb.
; Note that the first instructions of the routine can result is writes to the first page of the ROM area when
; there is no active invader bomb or laser base missile.
; The routine typically takes 1259 T-states and returns with a delay counter of $F3 to yield a delay of 3154 T-states, making
; a total of 4413 T-states.
; Entry: DE=Offset to the next row.
;        C=Inverse space.

L41AD:  LD      HL,($41F9)              ; 16    Fetch the laser base missile display file address.
        LD      (HL),C                  ; 7     Erase the missile (if there is no active missile then this writes to the first page of ROM).

        LD      HL,($41F7)              ; 16    Fetch the invader bomb display file address.
        LD      (HL),C                  ; 7     Erase the bomb (if there is no active bomb then this writes to the first page of ROM).

        LD      L,(IY+$22)              ; 19    Fetch the low address of the laser base display file address.
        LD      H,$42                   ; 7     HL points to the column above the laser base but on row 8 (the active last row of the invader area).

        ADD     HL,DE                   ; 11    Point to corresponding column in the display file address on row 9, i.e. below the invaders.

        LD      (IY+$21),D              ; 19    Signal that an invader has not yet been selected to drop a bomb.

; Enter a loop to test each row of the invader area to see if there is an invader directly overhead the laser base.

L41BE:  LD      A,(HL)                  ; 7     Fetch the character within the invader area above the laser base.
        CP      CH_INVADER_MIDDLE       ; 7     Is it the middle character of an invader? [$0B=ZX81, $01=ZX80]
        JR      NZ,L41D4                ; 12/7  Jump ahead if not.

; It is the middle of an invader.

        LD      A,($41F8)               ; 13    Fetch the high byte of the invader bomb display file position.
        AND     A                       ; 4     Is there already an active invader bomb?
        JR      NZ,L41D7                ; 12/7  Jump ahead if there is.

; There is not an active invader bomb.

        LD      A,($41FC)               ; 13    Has an invader been selected to drop a bomb?
        AND     A                       ; 4
        JR      NZ,L41DC                ; 12/7  Jump ahead if one has been selected.

; An invader has not yet been selected to drop a bomb.

        LD      ($41F7),HL              ; 16    Store the new invader bomb display file address.
        JR      L41E3                   ; 12    Jump ahead.

; The middle of an invader is not above the laser base.

L41D4:  PUSH    BC                      ; 11    Perform a delay to balance the alternate paths to $41D7.
        POP     IX                      ; 14

; Join here if an invader has already been selected to drop a bomb.

L41D7:  LD      A,(HL)                  ; 7     Does this location contain an invader?
        CP      $80                     ; 7
        JR      NZ,L41E0                ; 12/7  Jump ahead if it does.

; There is no invader at the current location.

L41DC:  JR      L41DE                   ; 12    Jump to the following instruction thereby performing a delay.

L41DE:  JR      L41E3                   ; 12    Jump ahead to continue.

; The current location contains an invader directly overhead the laser base and there is no active bomb.

L41E0:  LD      (IY+$21),E              ; 19    Signal that an invader has been selected to drop a bomb.

; Joins here when a new invader bomb has just been dropped.

L41E3:  AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point to the previous row within the invader area, but still directly above the laser base.

        LD      A,H                     ; 4
        CP      $41                     ; 7     Have all of the rows within the invader area been considered?
        JR      NZ,L41BE                ; 12/7  Jump back if not to test the previous row.

; All rows within the invader area have been tested to see whether an invader can drop a bomb.
; Make a return to the main loop, executing a delay upon return to pad out the processing time slot.

        LD      B,$F3                   ; 7     Set the delay loop counter
        NOP                             ; 4     Delay.
        JR      L41AB                   ; 12    Indirectly return to the main loop. The delay counter holds $F3, i.e. 243 iterations.

L41F0:

; ======================================================================================================================================================

; ================
; Transfer Routine
; ================
; If after loading, GOTO 100 followed by NEWLINE is immediately keyed in then the transfer routine runs from $4349 (1K) or $4649 16K.
; It is held in the variables area and hence will shift location based upon the available memory. This is because the ZX81 will produce
; a partially collapsed display file if there is insufficient memory. It's location is given by (VARS)+$1F6 = ($4010)+$1F6.
; The routine copies $01F0 bytes from $4159 (1K RAM) to $4000, and then runs the transferred program from $4000.
; The transfer routine occupies 20 bytes.

L4349:  LD      DE,L4000                ; The destination address.
        LD      HL,($4010)              ; Fetch the address of the variables area.
        INC     HL                      ;
        INC     HL                      ;
        INC     HL                      ;
        INC     HL                      ;
        INC     HL                      ;
        INC     HL                      ; Point to the payload for the first variable. It will be a string array containing the program bytes.
        LD      BC,L41F0-L4000          ;
        LDIR                            ; Copy the program bytes to their destination location.
        JP      L4000                   ; Run the transferred program.


; The ZX80 version could have been:
;
;L4289:  LD      DE,L4000               ; The destination address.
;        LD      HL,($4008)             ; Fetch the address of the variables area.
;        INC     HL                     ;
;        INC     HL                     ; Point to the payload for the first variable. It will be a numeric array containing the program bytes.
;        LD      BC,L41F0-L4000         ;
;        LDIR                           ; Copy the program bytes to their destination location.
;        JP      L4000                  ; Run the transferred program.
;
; The transfer routine occupies 16 bytes.

; ======================================================================================================================================================

; ============
; BASIC Loader
; ============
; Once the program bytes have been typed in, the program can be started with GOTO 100.
;
;  10 DIM A$(516)
;  20 FOR N=1 TO 516
;  30 CLS
;  40 PRINT "A$(";N;")=";CODE A$(N)
;  50 INPUT I$
;  60 IF I$="" THEN NEXT N
;  70 LET A$(N)=CHR$ VAL I$
;  80 GOTO 30
; 100 FAST
; 110 LET N=USR (PEEK 16400+256*PEEK 16401+502)


; The ZX80 version could have been:
;
;  10 DIM A(255)
;  20 FOR N=0 TO 255
;  30 CLS
;  40 PRINT "A(";N;")=";A(N)
;  50 INPUT I
;  60 LET A(N)=I
;  70 NEXT N
;  80 STOP
; 100 LET N=USR(PEEK(16392)+256*PEEK(16393)+498)

; ======================================================================================================================================================

; ===========
; FLOW CHARTS
; ===========

; -------------
; Move Invaders
; -------------
; $4096. The routine typically takes 3342 T-states.
;
;                               $4096                                MOVE ROW OF INVADERS
;                                 |
;                              15 |
;                                 v         12
;                                 +-----------------------+
;                               7 |   Move Invaders Left  |
;                                 v                       v
;                              22 |                       | 17
;                                 v                       |
;                           $40A5 +<----------------------+
;                                 |
;                               7 |
;                                 v
;         +---------------------->+ $40A6
;         |                       |
;         |                     8 |
;         |              |        v         12
;         |              |        +-----------------------+
;         |       30     |      7 |   Move Invaders Down  |
;         |   Iterations |        v                       v
;         |              |     53 |                       | 48
;         |              |        v                       |
;         |              v  $40BA +<----------------------+
;         |                       |
;         |                       | 25
;         |  End Row Not Reached  v
;         +-----------------------+
;                   12            | 7
;                                 v
;                                 | 70
;                                 v         12
;                                 +-----------------------+
;                               7 |   More Area To Shift  |
;                                 v                       |
;                              18 |                       |
;                   12            v                       v
;         +-----------------------+                       | 7
;         |  Not At Screen Edge   | 7                     v
;         |                       v                       |
;         |                       | 7                     |
;         |                 $40D8 v                       | 3*13-5
;         +---------------------->+                       |
;                                 | 16                    |
;                                 v                       |
;                           $40DF +<----------------------+
;                                 |
;                              11 |
;                                 v
;                              12 |
;                                 v
;                              To $407B


; --------------------------------------------
; Move the Invader Bomb and Laser Base Missile
; --------------------------------------------
; $40EF. The routine typically takes 1122 T-states.
;
;                               $40EF                                   MOVE INVADER BOMB
;                                 |
;                              39 |
;                   12            v
;         +-----------------------+
;         |      Active Bomb    7 |
;         |                       v
;         |                     7 |
;         |                       v         12
;         +---------------------->+-----------------------------------------------+
;                               7 |    Reached Row 16 or No Active Bomb           |
;                                 v                                               | 23
;                              11 |                                               |
;                                 v         12                              7     v
;                                 +------------------------------------>--------->+ $410C
;                               7 |    Hit Shield                                 |
;                                 v                                               |
;                               7 |                                               |
;                                 v         12                                    |
;                                 +-----------------------+                       | 25
;                               7 |    Hit Laser Base     | 42                    |
;                                 v                       |                       |
;                              10 |                       v                       |
;                                 v                   Restart Game                |
;                              12 |                     $4000                     |
;                                 v                                               |
;                           $410F +<----------------------------------------------+
;                                 |
;                                 |
;                              65 |                               MOVE LASER BASE MISSILE
;                   12            v
;         +-----------------------+
;         |                       |
;         |                     7 | Active Missile
;         |                       v
;         |                     7 |
;         |                       v         12
;         +---------------------->+-----------------------------------------------------------------------+
;                                 |    Off Screen                                                         |
;                               7 |                                                                       | 7
;                                 v                                                                       |
;                              14 |                                                                       v
;                                 v         12                                                            |
;                                 +-----------------------+                                               |
;                               7 |    Hit Invader Bomb   |                                               |
;                                 v                       | 19                                            | (13*4)-5
;                               4 |                       |                                               |
;                                 v         12            v                      19                       |
;                                 +-----------------------+---------------------------------------------->+
;                               7 |    Hit Shields                                                        |
;                                 v                                                                       |
;                               7 |                                                                       |
;                                 v         12                                                            |
;                                 +-----------------------+                                               |
;                               7 |    Hit Invader        | 15                                            |
;                                 v                       v                                               |
;                                 |                       | (24*5)-5                                      |
;                                 |                       v                                               |
;                                 |             Increment | 31                                            |
;                                 |             Score     v         12                                    |
;                                 |                       +-----------------------+                       |
;                                 |                     7 |   No 1st Digit Carry  |                       |
;                                 |                       v                       |                       |
;                              22 |                    11 |                       |                       |
;                                 |                       v                       |                       |
;                                 |                       +<----------------------+                       |
;                                 |                    39 |                                               |
;                                 |                       v         10                                    |
;                                 |                       +---------------------->+                       |
;                                 |                    10 |    Wave Destroyed     |                       |
;                                 |                       v                       v                       |
;                                 |                       |                    To $403E                   |
;                                 |                    12 |                                               |
;                                 |                       |                                               |
;                                 v          7            v                                               |
;                           $4137 +<----------------------+<----------------------------------------------+
;                                 |                     $4135
;                              23 |
;                                 v
;                       (13*63)-5 |
;                                 v
;                              12 |
;                                 v
;                              To $40E2


; -------------------------------
; Move and Display the Laser Base
; -------------------------------
; $4166. The routine typically takes 1119 T-states.
;
;                               $4166                                     MOVE LASER BASE
;                                 |
;                              71 |
;                                 v         12
;                                 +-----------------------------------------------+
;                               7 |                                               |
;               Key 8 Pressed     v                                               |
;                               7 |                                               |
;                                 v         12                                    v
;                                 +-----------------------+                       |
;                               7 |    Base At Column 28  |                       |
;                                 v                       |                       |
;                               4 |                       |                       | 26
;                                 v                       |                       |
;                           $417C +<----------------------+                       |
;                                 |                                               |
;                              12 |                                               |
;                                 v                                               |
;                           $4181 +<----------------------------------------------+
;                                 |
;                               8 |
;                                 v         12
;                                 +-----------------------------------------------+
;                               7 |                                               |
;               Key 5 Pressed     v                                               |
;                               7 |                                               |
;                                 v         12                                    v
;                                 +-----------------------+                       |
;                               7 |    Base At Column 2   |                       |
;                                 v                       |                       |
;                               4 |                       |                       | 26
;                                 v                       |                       |
;                           $418A +<----------------------+                       |
;                                 |                                               |
;                              12 |                                               |
;                                 v                                               |
;                           $418F +<----------------------------------------------+
;                                 |
;                              46 |
;                                 v         12
;                                 +-----------------------+
;                               7 |    No Active Missile  |
;                                 v                       v
;                               8 |                       | 14
;                                 v         12            |
;                                 +---------------------->+
;                               7 |                       |
;               Key 0 Pressed     v                       |
;                                 |                       | 21
;                              28 |                       |
;                                 v                       |
;                           $41A7 +<----------------------+
;                                 |
;                               7 |
;                                 v
;                                 | (7*13)-5
;                                 v
;                                 | 12
;                                 v
;                              To $413E


; -------------------------------------
; Check Whether to Drop an Invader Bomb
; -------------------------------------
; $41AD. The routine typically takes 1259 T-states.
;
;                               $41AD                      CHECK WHETHER TO DROP NEW BOMB
;                                 |
;                                 | 102
;                                 v
;         +---------------------->+ $41BE
;         |                       |
;         |                       | 14
;         |                       v         12
;         |                       +-----------------------+ $41D4
;         |                     7 |   Not Middle Chr$     |
;         |                       v                       | 25
;         |                    17 |                       |
;         |                       v         12            v
;         |                       +---------------------->+ $41D7
;         |                     7 |   Active Bomb Exists  |
;         |                       v                       | 14
;         |                       |                       v         12
;         |              |     17 |                       +---------------------->+
;         |              |        |                     7 |   Invader Not Found   |
;         |        9     |        v         12            v                       |
;         |   Iterations |        +---------------------->+ $41DC                 |
;         |              |      7 |   Bomb Already        |                       |
;         |              |        v   Selected            | 12                    | 19
;         |              v     16 |                       |                       |
;         |                       v                       v                       |
;         |    Select Invader  12 |                       |                       |
;         |    To Drop Bomb       v                       | 12                    |
;         |                 $41E3 +<----------------------+<----------------------+
;         |                       |
;         |      More Rows     30 |
;         |      To Check         v
;         +-----------------------+
;                   12            | 7
;                                 v
;                                 | 11
;                                 v
;                                 | 12
;                                 v
;                              To $41AB

; ======================================================================================================================================================

        END

